// MD-R
// version 3.0
// Sep 9 2015
// Edited by: Vi Q. Ha, Yihao Zhang

/// This is the three-dimension projection modification for Repast HPC

/*
 *   Repast for High Performance Computing (Repast HPC)
 *
 *   Copyright (c) 2010 Argonne National Laboratory
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with
 *   or without modification, are permitted provided that the following
 *   conditions are met:
 *
 *     Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *     Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 *     Neither the name of the Argonne National Laboratory nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 *   PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE TRUSTEES OR
 *   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *   EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 *  SharedGrid.cpp
 *
 *  Created on: Sep 11, 2009
 *      Author: nick
 */


#include "SharedBaseGrid3D.h"
#include <vector>
#include <boost/mpi.hpp>

using namespace std;

//~ namespace repast {
namespace vamde3d {

	using namespace repast;
Neighbors::Location sendDirections[26] = {               ///add other 18 neighbors by yihao
    Neighbors::NW, Neighbors::N, Neighbors::NE,
    Neighbors::W,                Neighbors::E,
    Neighbors::SW, Neighbors::S, Neighbors::SE,
    Neighbors::FNW, Neighbors::FN, Neighbors::FNE,
    Neighbors::FW,  Neighbors::F, Neighbors::FE,
    Neighbors::FSW, Neighbors::FS, Neighbors::FSE,
    Neighbors::BNW, Neighbors::BN, Neighbors::BNE,
    Neighbors::BW,  Neighbors::B, Neighbors::BE,
    Neighbors::BSW, Neighbors::BS, Neighbors::BSE
};

Neighbors::Location recvDirections[26] = { ///add other 18 neighbors by yihao
	
	Neighbors::BSE, Neighbors::BS, Neighbors::BSW,
	Neighbors::BE, Neighbors::B, Neighbors::BW,
	Neighbors::BNE, Neighbors::BN, Neighbors::BNW,
	Neighbors::FSE, Neighbors::FS, Neighbors::FSW,
	Neighbors::FE, Neighbors::F, Neighbors::FW,
	Neighbors::FNE, Neighbors::FN, Neighbors::FNW,
    Neighbors::SE, Neighbors::S, Neighbors::SW,
    Neighbors::E,                Neighbors::W,
    Neighbors::NE, Neighbors::N, Neighbors::NW
};


CartTopology::CartTopology(vector<int> processesPerDim, vector<double> origin, vector<double> extents, bool spaceIsPeriodic, boost::mpi::communicator* world) :
  periodic(spaceIsPeriodic), procsPerDim(processesPerDim) {
  
  /// // 2014.06.03 Vi's edit
  /// procsPerDim(processesPerDim) is the vector discribing how the processes are divided e.g {2,2,2}
  
  /// numDims discribes the number of dimensions in this topology e.g 2D or 3D
  int numDims = procsPerDim.size();
  
  /// an array of integers called periods. 
  /// I am guessing this is to store whether that particular dimension will be periodic
  int* periods = new int[numDims];
  
  /// the "smart" way of casting boolean periodic to integer periodicFlag
  int periodicFlag = periodic ? 1 : 0;
  
  /// assigning all components of periods to periodicFlag, 
  /// I am guessing this was designed to have different components pertain different value
  for (int i = 0; i < numDims; i++) periods[i] = periodicFlag;

	/// Why on earth would you do something like this, this is just begging for a disaster...:
	// swap xy so row major // Yihao said this is annoying:
	swapXY(processesPerDim);
	swapXY(procsPerDim);
	swapXY(origin);
	swapXY(extents);

	/// using *world, create an mpi cartesian... thing called topologyComm
	MPI_Cart_create(*world, numDims, &processesPerDim[0], periods, 0, &topologyComm);
	
	/// deleting the dynamic array
	delete[] periods;
	
	/// self explainatory:
	globalBounds = GridDimensions(origin, extents);
}

void CartTopology::getCoordinates(int rank, std::vector<int>& coords) {
	coords.assign(globalBounds.dimensionCount(), 0);
	MPI_Cart_coords(topologyComm, rank, globalBounds.dimensionCount(), &coords[0]);
}

GridDimensions CartTopology::getDimensions(int rank) {
	vector<int> coords;
	getCoordinates(rank, coords);
	return getDimensions(coords);
}


GridDimensions CartTopology::getDimensions(vector<int>& pCoordinates) {
  vector<double> origins, extents;
  for (size_t i = 0; i < pCoordinates.size(); i++) {
    double lower = globalBounds.origin(i) + ((double)pCoordinates[i] / (double)procsPerDim[i]) * globalBounds.extents(i);
    double upper = globalBounds.origin(i) + (((double)pCoordinates[i] + 1 )/ (double)procsPerDim[i]) * globalBounds.extents(i);
    origins.push_back(lower);
    extents.push_back(upper - lower);
  }

	swapXY(origins);
	swapXY(extents);
	return GridDimensions(Point<double> (origins), Point<double> (extents));
}


////////////
int CartTopology::getRank(vector<int>& loc, int rowAdj, int colAdj, int zAdj) { // added zAdj by Yihao // this function is to get rank number according to relative possition
	int* coord = new int[3];
	coord[0] = loc[0] + rowAdj;
	coord[1] = loc[1] + colAdj;
	coord[2] = loc[2] + zAdj;
	
	//////not modified because we use periodic boundary
	if (!periodic) {
		if (coord[0] < 0 || coord[0] > procsPerDim[0] - 1 || coord[1] < 0 || coord[1] > procsPerDim[1] - 1){
		  delete[] coord;
			return MPI_PROC_NULL;
		}
	}
	////
	
	int rank;
	MPI_Cart_rank(topologyComm, coord, &rank);
	delete[] coord;
	return rank;
}
//////////////

/// create neighbor according to relative posstions
void CartTopology::createNeighbors(Neighbors& nghs) {
  int eRank, wRank, nRank, sRank, bRank, fRank; //
  
	/// Synopsis: 
	/// int MPI_Cart_shift(MPI_Comm comm, int direction, int disp, int *rank_source, int *rank_dest)
	
	/// Why is this so weird??  
	MPI_Cart_shift(topologyComm, 1,  1, &wRank, &eRank);
	MPI_Cart_shift(topologyComm, 0, -1, &sRank, &nRank);
	MPI_Cart_shift(topologyComm, 2, -1, &bRank, &fRank);   //z direction // added by yihao

	createNeighbor(nghs, eRank, Neighbors::E);
	createNeighbor(nghs, wRank, Neighbors::W);
	createNeighbor(nghs, nRank, Neighbors::N);
	createNeighbor(nghs, sRank, Neighbors::S);
	createNeighbor(nghs, bRank, Neighbors::B);		// yihao
	createNeighbor(nghs, fRank, Neighbors::F);		// yihao
	

	int rank = RepastProcess::instance()->rank();
	vector<int> pCoordinates;
	getCoordinates(rank, pCoordinates);
	
////for 3D
	createNeighbor(nghs, getRank(pCoordinates, -1, -1, 0), Neighbors::NW);
	createNeighbor(nghs, getRank(pCoordinates, -1, 1, 0), Neighbors::NE);
	createNeighbor(nghs, getRank(pCoordinates, 1, -1, 0), Neighbors::SW);
	createNeighbor(nghs, getRank(pCoordinates, 1, 1, 0), Neighbors::SE);
	
	
	createNeighbor(nghs, getRank(pCoordinates, -1, -1, -1), Neighbors::FNW);
	createNeighbor(nghs, getRank(pCoordinates, -1, 1, -1), Neighbors::FNE);
	createNeighbor(nghs, getRank(pCoordinates, 1, -1, -1), Neighbors::FSW);
	createNeighbor(nghs, getRank(pCoordinates, 1, 1, -1), Neighbors::FSE);
	createNeighbor(nghs, getRank(pCoordinates, 0, -1, -1), Neighbors::FW);
	createNeighbor(nghs, getRank(pCoordinates, 0, 1, -1), Neighbors::FE);
	createNeighbor(nghs, getRank(pCoordinates, 1, 0, -1), Neighbors::FS);
	createNeighbor(nghs, getRank(pCoordinates, -1, 0, -1), Neighbors::FN);
	createNeighbor(nghs, getRank(pCoordinates, -1, -1, 1), Neighbors::BNW);
	createNeighbor(nghs, getRank(pCoordinates, -1, 1, 1), Neighbors::BNE);
	createNeighbor(nghs, getRank(pCoordinates, 1, -1, 1), Neighbors::BSW);
	createNeighbor(nghs, getRank(pCoordinates, 1, 1, 1), Neighbors::BSE);
	createNeighbor(nghs, getRank(pCoordinates, 0, -1, 1), Neighbors::BW);
	createNeighbor(nghs, getRank(pCoordinates, 0, 1, 1), Neighbors::BE);
	createNeighbor(nghs, getRank(pCoordinates, 1, 0, 1), Neighbors::BS);
	createNeighbor(nghs, getRank(pCoordinates, -1, 0, 1), Neighbors::BN);				
////	
	
}

void CartTopology::createNeighbor(Neighbors& nghs, int rank, Neighbors::Location location) {
	if (rank != MPI_PROC_NULL) {
		Neighbor* ngh = new Neighbor(rank, getDimensions(rank));
		nghs.addNeighbor(ngh, location);
	}
}

Neighbor::Neighbor(int rank, GridDimensions bounds) :
	_rank(rank), _bounds(bounds) {
}


Neighbors::Neighbors() {
	for (int i = 0; i < Neighbors::LOCATION_SIZE; i++) {
		nghs.push_back(0);
	}
}

void Neighbors::addNeighbor(Neighbor* ngh, Neighbors::Location location) {
	nghs[location] = ngh;
}

Neighbor* Neighbors::neighbor(Neighbors::Location location) const {
	return nghs[location];
}

Neighbor* Neighbors::findNeighbor(const std::vector<int>& pt) {
	for (std::vector<Neighbor*>::iterator iter = nghs.begin(); iter != nghs.end(); ++iter) {
		Neighbor* ngh = *iter;
		if ((ngh != 0) && (ngh->bounds().contains(pt)))
			return ngh;
	}
	return 0;
}

Neighbor* Neighbors::findNeighbor(const std::vector<double>& pt) {
  for (std::vector<Neighbor*>::iterator iter = nghs.begin(); iter != nghs.end(); ++iter) {
		Neighbor* ngh = *iter;
		if ((ngh != 0) && (ngh->bounds().contains(pt)))
			return ngh;
	}
	return 0;
}

Neighbors::~Neighbors() {
	for (vector<Neighbor*>::iterator iter = nghs.begin(); iter != nghs.end(); iter++) {
		Neighbor* ngh = *iter;
		delete ngh;
	}
}

ostream& operator<<(ostream& os, const Neighbors& nghs) {
	Neighbor* ngh = nghs.neighbor(Neighbors::NW);
	if (ngh != 0) {
		os << "\tNW ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::NE);
	if (ngh != 0) {
		os << "\tNE ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::SW);
	if (ngh != 0) {
		os << "\tSW ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::SE);
	if (ngh != 0) {
		os << "\tSE ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::N);
	if (ngh != 0) {
		os << "\tN ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::S);
	if (ngh != 0) {
		os << "\tS ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::E);
	if (ngh != 0) {
		os << "\tE ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	ngh = nghs.neighbor(Neighbors::W);
	if (ngh != 0) {
		os << "\tW ngh: " << ngh->rank() << ": " << ngh->bounds() << "\n";
	}

	return os;

}

}
