/* Agent.h */

#ifndef AGENT
#define AGENT

#include "AgentId.h"
#include "SharedContext.h"
#include "Grid.h"
#include "src.h"

/* Agents */
class Agent{
	
private:
    repast::AgentId   id_;
    Vect   dis_;         //displacement
    Vect   vel_;         //velocity
    Vect   acc_;         //acceleration
    Vect   force_;       //force acting on the agent
    Vect   updis;        //temporary displacement used in leapfrog or other methods
    Vect   upvel;        //temporary velocity used in leapfrog or other methods
    Vect   upacc;        //temporary acceleration used in leapfrog or other methods
    double mass;         //mass of agent
	std::vector<double> R;       //vector used to store the displacements for moving
	
	
public:
    
	Agent(){}
    Agent(repast::AgentId id, Vect dis, Vect vel, Vect acc, Vect updis_, Vect upvel_, Vect upacc_);
   
	
    ~Agent();
    
    /* Required Getters */
    virtual repast::AgentId& getId(){                   return id_;    }
    virtual const repast::AgentId& getId() const {      return id_;    }
	
   
    Vect getdis();             
	Vect getvel();	              
    Vect getacc(); 
    Vect getupdis();             
	Vect getupvel(); 
    Vect getupacc();
    
    double getmass();
    
    void initforce();          // force to 0
    void addforce(Vect force);   //accumulate force acting on the agent 
    void computupacc();          // compute acc with accumulated force      	
	void setR();              //give the value of displacement in each direction
		                       
	
	
    //~ /* Setter */
    void set(int currentRank, Vect dis, Vect vel, Vect acc, Vect updis_, Vect upvel_, Vect upacc_);
    void setdis(Vect dis);
    void setvel(Vect vel);
    void setacc(Vect acc);
    void setupdis(Vect dis);
    void setupvel(Vect vel);
    void setupacc(Vect acc);

    void update();              //update dis vel acc to t+timestep
	std::pair<bool, repast::Point<double> > move(repast::Grid<Agent, double>* grid);         //move agent by displacement
    /* Actions */
    //~ bool cooperate();                                                 // Will indicate whether the agent cooperates or not; probability determined by = c / total
    //~ void play(repast::SharedContext<Agent>* context);    // Choose three other agents from the given context and see if they cooperate or not
	
};
//~ 
/* Serializable Agent Package */
struct AgentPackage {
	
public:
    int    id;
    int    rank;
    int    type;
    int    currentRank;
    Vect   dis;
    Vect   acc;
    Vect   vel;
	Vect   updis;
    Vect   upvel;
    Vect   upacc;
    
    
    /* Constructors */
    AgentPackage(); // For serialization
    AgentPackage(int _id, int _rank, int _type, int _currentRank,Vect _dis, Vect _vel, Vect _acc,Vect _updis, Vect _upvel, Vect _upacc);
	
    /* For archive packaging */
    template<class Archive>
    void serialize(Archive &ar, const unsigned int version){
        ar & id;
        ar & rank;
        ar & type;
        ar & currentRank;
        ar &   dis.x;
        ar &   dis.y;
        ar &   dis.z;
        ar &   vel.x;
        ar &   vel.y;
        ar &   vel.z;
        ar &   acc.x;
        ar &   acc.y;
        ar &   acc.z;
 	    ar &   updis.x;
 	    ar &   updis.y;
 	    ar &   updis.z;
        ar &   upvel.x;
        ar &   upvel.y;
        ar &   upvel.z;
        ar &   upacc.x;
        ar &   upacc.y;
        ar &   upacc.z;
       
    }
    
	//~ // Needed for Shared Network
	//~ repast::AgentId getId() const {
 	//~ return repast::AgentId(id, rank, type);
    //~ }
	
};


#endif
