/* Model.h */

#ifndef MODEL
#define MODEL

#include <boost/mpi.hpp>
#include "Schedule.h"
#include "Properties.h"
#include "SharedContext.h"
#include "AgentRequest.h"
#include "SharedNetwork.h"
#include "SharedContinuousSpace3D.h"   //change from original "SharedContinuousSpace.h" in include/repast_hpc to this
#include "GridComponents.h"


#include "Agent.h"
#include "src.h"


/* Agent Package Provider */
class AgentPackageProvider {
	
private:
    repast::SharedContext<Agent>* agents;
	
public:
	
    AgentPackageProvider(repast::SharedContext<Agent>* agentPtr);
	
    void providePackage(Agent * agent, std::vector<AgentPackage>& out);
	
    void provideContent(repast::AgentRequest req, std::vector<AgentPackage>& out);
	
};

/* Agent Package Receiver */
class AgentPackageReceiver {
	
private:
    repast::SharedContext<Agent>* agents;
	
public:
	
    AgentPackageReceiver(repast::SharedContext<Agent>* agentPtr);
	
    Agent * createAgent(AgentPackage package);
	
    void updateAgent(AgentPackage package);
	
};


class Model{
	int stopAt;
	int countOfAgents;
	int PrNx, PrNy, PrNz, PrNtotal;
	int buffer;
	double timestep;
	double rcut;
	double dxij, dyij, dzij, dxy, redxy, Fij, mass;
	repast::Properties* props;
	repast::SharedContext<Agent> context;

	mdr3d::SharedContinuousSpace3D<Agent, repast::WrapAroundBorders, repast::SimpleAdder<Agent> >* continuousSpace;
	//~ repast::SharedContinuousSpace<Agent, repast::StrictBorders, repast::SimpleAdder<Agent> >* continuousSpace;
    
    Vect force;       
    Vect dis;
    Vect vel;
    Vect acc;
    Vect updis;
    Vect upvel;
    Vect upacc;
    
    
    std::vector<double> agentLoci;
    std::vector<double> agentLocj;
    Vect  Loci, Locj;
    Vect  agentVeli, agentAcci;
    Vect  agentVelj, agentAccj;
    
    struct acrossP
    {  
 
    Vect v;
    Vect a;
		
		}Across[500][5][1];    //used to store the values of dis vel and acc of agents 
    
	AgentPackageProvider* provider;
	AgentPackageReceiver* receiver;
	
public:
	Model(std::string propsFile, int argc, char** argv, boost::mpi::communicator* comm);
	~Model();
	
	
	
	
	
	void init();
	//~ void requestAgents();
	//~ void cancelAgentRequests();
	//~ void removeLocalAgents();
	//~ void connectAgentNetwork();
	void doSomething();
	void Interaction();
	void initSchedule(repast::ScheduleRunner& runner);
	void recordResults();
	void writeCFG();            //write to cfg for processor 0
	void writeothers();          //write to cfg for other processors
};

#endif
